#ifndef NEMO_MESS822_H
#define NEMO_MESS822_H

#include <nemo/stdint.h>
#include <nemo/unixtypes.h>
#include <nemo/stralloc.h>
#include <nemo/sa_vector.h>
#include <nemo/tai.h>
#include <nemo/caldate.h>
#include <nemo/caltime.h>
#include <nemo/djbio.h>

#ifdef __cplusplus
extern "C" {
#endif

#define MESS822_MAX_LINE_LENGTH 1024

typedef struct {
  struct caltime ct;
  unsigned int known;  /* 0 for ct uninitialized; 1 if ok; 2 if ok and right zone */
} mess822_time;

unsigned int	mess822_date(const mess822_time *when, stralloc *out);
unsigned int	mess822_when(mess822_time *out, const stralloc *in);
void		mess822_tai(mess822_time *out, const struct tai *in);


typedef struct {
  stralloc data;
  char type;  /* '@'=address, '('=comment, ':'=group name, ';'=group end */
} mess822_addrdata;

#define MESS822_ADDRDATA {STRALLOC,0}

typedef struct {
  mess822_addrdata *ad;
  unsigned int len;
  unsigned int a;
} mess822_addrlist;

#define MESS822_ADDRLIST {0,0,0}

static inline void mess822_addrlist_init(register mess822_addrlist *al)
{
  al->ad = 0;
  al->a = al->len = 0;
}

/*  return: 1 = success, 0 = fail  */
unsigned int mess822_addrlist_realloc(mess822_addrlist *v, unsigned int n);
unsigned int mess822_addrlist_alloc(mess822_addrlist *v, unsigned int n);

static inline unsigned int mess822_addrlist_ready(register mess822_addrlist *al, register unsigned int n)
{
  if (al->ad) {
    if (n <= al->a) return 1;
    return mess822_addrlist_realloc(al, n);
  }
  return mess822_addrlist_alloc(al, n);
}

static inline unsigned int mess822_addrlist_readyplus(register mess822_addrlist *al, register unsigned int n)
{
  if (al->ad) {
    n += al->len;
    if (n <= al->a) return 1;
    return mess822_addrlist_realloc(al, n);
  }
  return mess822_addrlist_alloc(al, n);
}

unsigned int	mess822_addrlist_scan(mess822_addrlist *out, const stralloc *in);
unsigned int	mess822_addrlist_cat(mess822_addrlist *to, const mess822_addrlist *from);
unsigned int	mess822_addrlist_copy(mess822_addrlist *to, const mess822_addrlist *from);
unsigned int	mess822_addrlist_erase(mess822_addrlist *addrlist);
unsigned int	mess822_addrlist_append(mess822_addrlist *addrlist, const stralloc *data, char type);
/*
  return length, safely
*/
static inline unsigned int mess822_addrlist_len(register const mess822_addrlist *v)
{
  if (v->ad) return v->len;
  return 0;
}


typedef struct {
  const char *name; /* 0 means all names */
  unsigned int *flag;
  sa_vector *copy;
  sa_vector *value;
  mess822_addrlist *addr;
  mess822_time *when;
} mess822_action;

#define MESS822_ACTION { 0, 0, 0, 0, 0, 0 }

typedef struct {
  stralloc inprogress;
  mess822_action *action;
  unsigned int (*exclude)(const stralloc *sa);
} mess822_header;

#define MESS822_HEADER { STRALLOC, 0, 0 }

unsigned int	mess822_begin(mess822_header *h, mess822_action *a);
unsigned int	mess822_line(mess822_header *h, const stralloc *s);
unsigned int	mess822_end(mess822_header *h);
unsigned int	mess822_slurp_headers(djbio *in, stralloc *inbuff, mess822_action *a);
void		mess822_header_excludes(mess822_header *h, const sa_vector *list);  /* list entries must be null terminated */

/* general routines */

unsigned int	mess822_ok(const stralloc *sa);

unsigned int	mess822_encodeplus(stralloc *out, const stralloc *data);
unsigned int	mess822_encode(stralloc *out, const stralloc *data);
unsigned int	mess822_quoteplus(stralloc *out, const stralloc *addr, const stralloc *comment);
unsigned int	mess822_quote(stralloc *out, const stralloc *addr, const stralloc *comment);
unsigned int	mess822_quotelist(stralloc *out, const mess822_addrlist *list);
unsigned int	mess822_encode_data(stralloc *out, const char *buf, unsigned int len, const char *pre, const char *post);
unsigned int	mess822_encode_stralloc(stralloc *out, const stralloc *sa, const char *pre, const char *post);
unsigned int	mess822_need_quote(const stralloc *sa, unsigned int len);

unsigned int	mess822_fold(stralloc *out, const stralloc *in, const char *prefix, unsigned int linelen);

unsigned int	mess822_token(sa_vector *out, const stralloc *in);

unsigned int	mess822_key_value_pairs(sa_vector *keys, sa_vector *values, const stralloc *in);

unsigned int	mess822_content_field(stralloc *rvalue, sa_vector *keys, sa_vector *values, const stralloc *in);

unsigned int	mess822_mime_delimiter(stralloc *out, const char *prefix);
unsigned int	mess822_message_id_line(stralloc *out, const char *box, const stralloc *host);
unsigned int	mess822_multipart_line(stralloc *out, const char *type, const stralloc *delimiter);

#ifdef __cplusplus
}
#endif

#endif
