#include <nemo/byte.h>
#include <nemo/str.h>

#include "mess822.h"

static unsigned int append_until(stralloc *out, const stralloc *in, unsigned int *pos, const char *until_chars)
{
  unsigned int uclen;
  unsigned int escape;
  unsigned int i;
  char ch;

  i = *pos;
  escape = 0;
  uclen = str_len(until_chars);
  while (i < in->len) {
    ch = in->s[i];
    if (escape) {
      escape = 0;
    }
    else {
      if (ch == '\\') {
	escape = 1;
	i++;
	continue;
      }
      if (byte_chr(until_chars, uclen, ch) != uclen) break;
    }
    if (!stralloc_append(out, &ch)) return 0;
    i++;
  }
  *pos = i;
  return 1;
}

static unsigned int append_comment(stralloc *out, const stralloc *in, unsigned int *pos)
{
  unsigned int escape;
  unsigned int level;
  unsigned int i;
  char ch;

  i = *pos;
  escape = 0;
  level = 1;
  while (i < in->len) {
    ch = in->s[i++];
    if (escape) {
      escape = 0;
    }
    else {
      if (ch == '\\') {
	escape = 1;
	continue;
      }
      else if (ch == '(') {
	level++;
      }
      else if (ch == ')') {
        if (level == 1) break;
        level--;
      }
    }
    if (!stralloc_append(out, &ch)) return 0;
  }
  *pos = i;
  return 1;
}

unsigned int mess822_token(sa_vector *out, const stralloc *in)
{
  static stralloc buffer = STRALLOC;

  unsigned int i;
  char ch;

  if (!sa_vector_erase(out)) return 0;

  i = 0;
  while (i < in->len) {
    switch (ch = in->s[i]) {
      case '"':
        i++;  /* skip over current char */
        if (!stralloc_copyb(&buffer, "%", 1)) return 0;
        if (!append_until(&buffer, in, &i, "\"")) return 0;
        if (!sa_vector_append(out, &buffer)) return 0;
        i++;  /* skip over '"' char */
        break;

      case '[':
        i++;  /* skip over current char */
        if (!stralloc_copyb(&buffer, "%[", 2)) return 0;
        if (!append_until(&buffer, in, &i, "]")) return 0;
        if (!stralloc_append(&buffer, "]")) return 0;
        if (!sa_vector_append(out, &buffer)) return 0;
        i++;  /* skip over ']' char */
        break;

      case '(':
        i++;  /* skip over current char */
        if (!stralloc_copyb(&buffer, "(", 1)) return 0;
        if (!append_comment(&buffer, in, &i)) return 0;
        if (!sa_vector_append(out, &buffer)) return 0;
        break;

      case '<': case '>':
      case ',': case ';': case ':': case '=':
      case '@': case '.':
      case ' ': case '\t':
        if (!stralloc_erase(&buffer)) return 0;
        if (!stralloc_append(&buffer, &ch)) return 0;
        if (!sa_vector_append(out, &buffer)) return 0;
        i++;  /* skip over current char */
        break;

      default:
        if (!stralloc_copyb(&buffer, "%", 1)) return 0;
        if (!append_until(&buffer, in, &i, "\"[(<>,;:=@. \t")) return 0;
        if (!sa_vector_append(out, &buffer)) return 0;
        break;
    }
  }
  return 1;
}
